/*

  Compiler settings for MSVC 4.2b:

  (suggested)

  preprocessory defs:  NDEBUG ( not  _DEBUG )
  calling convention: fastcall
  processor: pentium
 
   optimizations:
      Assume No Aliasing
      Global Optimizations 
      Generate Intrinsic Functions
      Favor Fast Code 
      Frame-Pointer Omission 
      Full Optimization

  In-line Function Expansion
    Only __inline

  not using precompiled headers
  not incrementally linked or compiled

*/

#include "first.h"
#include "raster_util.h"
#include "clut.h"
#include "raster.h"
#include <assert.h>




//
// color range of palletes stored with textures
// this is 0..255 for .bmps.
//
const float src_pallette_color_range = 255.0;

 
//
// for statically allocated buffers
//
const int max_shade_bits  = 8;
const int max_shade_range = 1<<max_shade_bits;


using namespace clut_miser;


int generator::calculate_clut_size(const color_table & table, int shade_bits ) const 
{ 
    return  (bytes_per_pixel * table.num_colors ) <<  
                 ((shade_bits == -1 ) ? settings.default_shade_bits : shade_bits );
}

u32 generator::suggest_clut_alignment(int shade_bits) const
{
 u32 required =  u32(bytes_per_pixel) << (shade_bits  == -1 ? settings.default_shade_bits : shade_bits);
 if( required >= 32ul )
   return  required;
 else 
   return  32ul;
}




output_settings::output_settings() :
default_shade_bits(6),
diffuse_scale(1.25f),
specular_start(0.75f),
specular_scale(1.0f)
{
}


generator::generator( int _bytes_per_pixel) : 
bytes_per_pixel( _bytes_per_pixel ) {}


generator_8::generator_8(const color_table & _video_palette) :
generator(1),
video_palette(_video_palette)
{}

generator_16::generator_16( const pixel_format_16 & _format ) :
generator(2),
format(_format)
{}

 // converts from 8,8,8 to 16 bit value

u32 generator_16::color_match(u32 r, u32 g, u32 b) const
{
  u32 rv = ( ( (r>>format.r_size_shift)  << (format.r_place_shift)  ) | 
             ( (g>>format.g_size_shift)  << (format.g_place_shift)  ) | 
             ( (b>>format.b_size_shift)  << (format.b_place_shift)  ) );
  return rv | (rv<<16);
}

u32 generator_8::color_match(u32 r, u32 g, u32 b) const 
{
   assert( video_palette.num_colors );
   
   //
   // uses an L2 norm metric.
   // this could be substituted for any other method
   // you could even let the win32 sdk do it for you.
   //

   float min_err = 1e15f;
   int min_idx=0;
   for(int h=0;h<video_palette.num_colors;++h)
   {
    float err = square( (float)( r-video_palette.entries[h][2]) ) +
                square( (float)( g-video_palette.entries[h][1]) ) +
                square( (float)( b-video_palette.entries[h][0]) );

    if( err < min_err )
    {
     min_err = err;
     min_idx = h;
    }
   } 

   return min_idx + (min_idx<<8) + (min_idx<<16) + (min_idx<<24);
}  






//
// linear coeffs for calculating clut.
//
struct coef_table
{
 int  num_shades;
 float coefs[max_shade_range];   // linear coefs
 float  offs[max_shade_range];   // constant offsets
  
 void init(const output_settings & settings,int shades);
};


//
// makes an entry in the clut for a single color with num_shades shades
//
static void make_single_clut_color(const generator & gen, void * out,float rgbs[3],const coef_table &);



void generator::make_clut(clut_t clut,const color_table & table, int _shade_bits) const 
{
   assert(clut );
    
   int shade_bits = _shade_bits != -1 ?  _shade_bits : settings.default_shade_bits; 

   coef_table coefs;
   int shades = 1 << shade_bits;

   // make coefs
   coefs.init(settings,shades);

   // allocate clut
   void * put = clut;
   int bytes_per_color = shades * bytes_per_pixel;

   for(int pidx=0;pidx<table.num_colors;pidx++, (int&)put += bytes_per_color )
   {
     float rgbs[3];
     for(int c=0;c<3;++c)
      {
       rgbs[c] = table.entries[pidx][c] / src_pallette_color_range;
      }
     make_single_clut_color(*this, put,rgbs,coefs);
   }

 }





pixel_format_16::pixel_format_16(int red_bits, int green_bits, int blue_bits,
                                 int r_place, int g_place, int b_place )
{
  r_size_shift = 8-red_bits; g_size_shift = 8-green_bits;  b_size_shift = 8-blue_bits;
  if( (r_place == 0)&&(g_place == 0)&&(b_place == 0) )
  {
    r_place_shift = blue_bits + green_bits;
    g_place_shift = blue_bits;
    b_place_shift = 0;
  }
  else
  {
    r_place_shift = r_place;
    g_place_shift = g_place;
    b_place_shift = b_place;
  }
 }


/*void clut_miser::init_8_bit(const color_table & _video_palette )
 {
  bytes_per_pixel = 1;
  set_video_palette(  _video_palette );
 }*/


void color_table::swap_r_b()
 {
  for(int h=0;h<num_colors;++h)
    {
      swap_obs(entries[h][0],entries[h][2]);
    } 
 }


void coef_table::init(const output_settings & settings, int shades)
{
   num_shades = shades;

   float shade_step = 1.0/( shades-1);

    // scale induced by specular level not being 1.0

   float scaled_specular_scale;
   float scaled_diffuse_scale;
   // any specular?
   if( settings.specular_start < 1.0f )
   {
    scaled_specular_scale = settings.specular_scale/(1.0f-settings.specular_start);
    scaled_diffuse_scale  = settings.diffuse_scale/(settings.specular_start);
   }
   else
   {
    scaled_specular_scale = 0;
    scaled_diffuse_scale  = settings.diffuse_scale;
   }


   for(int es=0;es< shades ;es++)
   {
    float shade =  es*shade_step;
    
 
    if( shade >= settings.specular_start )
     {
      offs[es] = (shade - settings.specular_start )*scaled_specular_scale;
      shade = settings.specular_start;
    }
    else
     {
      offs[es] = 0;
     }
       // linear
    coefs[es] = shade*scaled_diffuse_scale;
 
   } 
}


pixel_blend_table_8 generator_8::create_blend_table() const
{
  pixel_blend_table_8 rv;

  assert( video_palette.num_colors == 256);

  // make blend table;
   for(int i=0;i<256;++i)
   {
    rv.maps[ i ][  i ] = i;
    for(int j=i+1;j<256;++j)
    {
     // 50,50 blend
     int r = (video_palette.entries[i][2] + video_palette.entries[j][2])>>1;
     int g = (video_palette.entries[i][1] + video_palette.entries[j][1])>>1;
     int b = (video_palette.entries[i][0] + video_palette.entries[j][0])>>1;

     int pix = color_match(r,g,b);

     rv.maps[i][j] =  pix;
     rv.maps[j][i] =  pix;

    }
   } 
   return rv;
}










static void make_single_clut_color(const generator & gen,
                                   void * store,
                                   float rgbs[3],
                                   const coef_table & coefs)
 {
   int out[3]; // r,g,b

   int bytes_per_pixel = gen.bytes_per_pixel;

   for(int shade=0;shade< coefs.num_shades; ++shade)
   {
	
	    for(int c=0;c<3;++c)
	    {
         float val = rgbs[c]*coefs.coefs[shade] + coefs.offs[shade];
 	    
         out[c] =  f_to_int_s( val, 8 );

         if( out[c] < 0 )
          out[c] = 0;
         if( out[c] > 255 )
          out[c] = 255;
     }
     if( bytes_per_pixel == 1)
     {
       *(byte *)store = gen.color_match(out[2],out[1],out[0]);
     }
     else
     {
       *(u16 *)store = gen.color_match(out[2],out[1],out[0]);
     }
     (int&)store += bytes_per_pixel;


   }
 }



